
import sys, glob, struct, os

import ait_data, fpga_data
from fpga_data import *

NUM_BLOCKS = LOCAL_MEMORY_SIZE / BLOCK_SIZE
B_CACHE = max(CS1, CS2, BRT_CACHE)
B_SPM = max(CS1 + fpga_data.Save(NUM_BLOCKS) + fpga_data.Load(NUM_BLOCKS),
        fpga_data.Restore(NUM_BLOCKS) + CS2, BRT_SPM)

(CALL, JUMP, RETURN, BRANCH, UNSUPPORTED) = "cjrbu"
CONDITIONS = set(["eq", "ne", "cs", "cc", "mi", "pl",
            "vs", "vc", "hi", "ls", "ge", "lt", "gt", "le"])
NAME_FIXUP = { 
                "bs":"binarysearch",
                "qsort-exam":"qsortexam",
            }

def analyse_ecb_ucb(prog_name):
    functions = dict()
    names = dict()
    instructions = dict()
    branches = dict()
    objdump = "objdumps/%s.od" % prog_name

    print ''
    print 'analysing program %s' % prog_name

    for line in file(objdump):
        fields = line.split()
        if len(fields) < 2:
            continue

        if fields[0].endswith(":"):
            fields[0] = fields[0][:-1]

        try:
            address = int(fields[0], 16)
        except:
            continue

        if fields[1].endswith(">:"):
            assert fields[1].startswith("<")
            name = fields[1][1:-2]
            functions[name] = address
            names[address] = name
            continue
        
        if (len(fields[1]) != 8) or (len(fields) < 3):
            continue

        try:
            iword = int(fields[1], 16)
        except:
            continue

        if fields[2][0] == '.':
            continue

        mnem = fields[2]
        conditional = False
        if mnem[-2:] in CONDITIONS:
            mnem = mnem[:-2]
            conditional = True

        instructions[address] = line.rstrip().replace("\t", "  ")
        inst = ' '.join(fields[2:])
        itype = None
        if mnem == "bl":
            itype = CALL

        elif mnem == "b":
            itype = JUMP

        elif (mnem == "bx") and (fields[3] == "lr"):
            itype = RETURN

        elif ((inst.startswith("ldm sp") or mnem == "pop")
        and (inst.find("pc}") > 0)):
            itype = RETURN

        elif inst.startswith("bic"):
            itype = None # Not a branch

        elif inst.startswith("b"):
            itype = UNSUPPORTED


        if itype != None:
            target = None
            if (itype != RETURN) and (itype != UNSUPPORTED):
                target = int(fields[3], 16)

            branches[address] = (itype, conditional, target)


    print '%u functions, %u instructions, %u branches' % (len(functions), 
                len(instructions), len(branches))

    trace2 = file("traces/%s.bt" % prog_name, "rb")

    whitham = set()
    executed = dict()
    counter = dict()
    old_address = 0
    word = trace2.read(4)
    while len(word) == 4:
        (address, ) = struct.unpack("<I", word)
        line = (address % LOCAL_MEMORY_SIZE) / BLOCK_SIZE
        whitham.add(line)
        executed[address] = old_address
        counter[address] = counter.get(address, 0) + 1
        old_address = address 

        word = trace2.read(4)


    linear_section = set()
    for (address, count) in counter.iteritems():
        if count == 1:
            linear_section.add(address)

    trace2.seek(0, 0)
    region_starts = []
    word = trace2.read(4)
    while len(word) == 4:
        (address, ) = struct.unpack("<I", word)
        if address in linear_section:
            region_starts.append((address, True))
            was_linear = True
        elif was_linear:
            region_starts.append((address, False))
            was_linear = False
             
        word = trace2.read(4)

    print '%u executed instructions, %u linear instructions' % (
                len(executed), len(linear_section))

    done = explore_program(instructions, 
                    branches, functions["main"], None)
    ecb = set([ (x % LOCAL_MEMORY_SIZE) / BLOCK_SIZE for x in done ])

    print '|ECB| = %u' % (len(ecb))

    altmeyer = set()
    for line in file("ecb-%ukb.txt" % (LOCAL_MEMORY_SIZE / 1024)):
        fields = line.split()
        name = fields.pop(0)

        name = NAME_FIXUP.get(name, name)

        if name == prog_name:
            assert len(fields) == (LOCAL_MEMORY_SIZE / BLOCK_SIZE)
            altmeyer.clear()
            for (i, x) in enumerate(fields):
                assert x in "01"
                if x == "1":
                    altmeyer.add(i)


    assert len(altmeyer) != 0, "Did not find %s in ECB file" % prog_name

    fd = file("analysis/%s.idump" % prog_name, "wt")
    done = list(done)
    done.sort()
    fd.write('ECB set size %u\n' % len(ecb))
    fd.write('measurement set size %u\n' % len(whitham))
    fd.write('Altmeyer discrepancy %d\n' % (len(ecb) - len(altmeyer)))

    fd.write("ecb   %s\n" % ecb_set_to_string(ecb))
    fd.write("altm  %s\n" % ecb_set_to_string(altmeyer))
    fd.write("run   %s\n" % ecb_set_to_string(whitham))

    errors = whitham - ecb
    if len(errors) != 0:
        executed = executed.items()
        executed.sort()
        fd.write("discrepancies:\n\n")
        for (address, prev) in executed:
            line = (address % LOCAL_MEMORY_SIZE) / BLOCK_SIZE
            if line in errors:
                fd.write(instructions[address])
                fd.write("  from %x\n" % prev)

        print len(errors), 'ERRORS DETECTED'
        sys.exit(1)

    # discover region contents
    regions = []
    was_linear = False
    enter_non_linear = None
    for (address, linear) in region_starts:
        if linear and was_linear:
            # merge
            assert (len(regions) % 2) == 1 # odd number of regions
            regions[-1].add(address)
            assert enter_non_linear == None
        
        elif linear and not was_linear:
            # enter linear region; find extents of previous region first
            if enter_non_linear != None:
                contents = explore_program(instructions, 
                        branches, enter_non_linear, address)
                regions.append(contents)
                enter_non_linear = None

            # create new linear region
            regions.append(set())
            assert (len(regions) % 2) == 1 # odd number of regions
            regions[-1].add(address)

        elif (not linear) and was_linear:
            # leave linear region
            assert enter_non_linear == None
            enter_non_linear = address

        elif (not linear) and (not was_linear):
            assert enter_non_linear != None

        else:
            assert False

        was_linear = linear


    fd.write('\n\n')
    num_ucbs = 0
    for i in xrange(len(regions)):
        ucb = set([ (x % LOCAL_MEMORY_SIZE) / BLOCK_SIZE for x in regions[i] ])
        fd.write('UCB%-2u %s\n' % (i, ecb_set_to_string(ucb)))
        num_ucbs = max(num_ucbs, len(ucb))
        assert len(ucb - ecb) == 0

    fd.write('UCB set size %u\n' % num_ucbs)
    print '|UCB| = %u' % num_ucbs

    # merge regions that share contents
    assert (len(regions) % 2) == 1 # odd number of regions
    for i in xrange(len(regions)):
        for j in xrange(i + 1, len(regions)):
            shared = regions[i] & regions[j]
            if len(shared) != 0:
                for k in xrange(i + 1, j + 1):
                    regions[i] |= regions[j]
                    regions[j] = set()
    i = 0
    while i < len(regions):
        if len(regions[i]) == 0:
            regions.pop(i)
        else:
            i += 1

    print len(regions), 'pre-merge regions'

    # merge regions according to the maximum size rule
    biggest = 0
    for i in xrange(len(regions)):
        biggest = max(len(regions[i]), biggest)

    i = 1
    while i < len(regions):
        if (len(regions[i-1]) + len(regions[i])) <= biggest:
            regions[i-1] |= regions[i]
            regions.pop(i)
        else:
            i += 1


    print len(regions), 'post-merge regions'

    for i in xrange(len(regions)):
        fd.write("region %u: %u instructions\n" % (i + 1, len(regions[i])))

    fd.write("\n")
    fd.write("\n")

    for i in xrange(len(regions)):
        fd.write("region %u of %u\n" % (i + 1, len(regions)))
        contents = list(regions[i])
        contents.sort()
        for address in contents:
            fd.write(instructions[address])
            fd.write("\n")

        fd.write("\n")

    old_address = 0

    for address in done:
        if address != (old_address + 4):
            fd.write('\n')

        if names.has_key(address):
            fd.write('\n%08x <%s>:\n' % (address, names[address]))

        fd.write(instructions[address])
        x = len(instructions[address])
        if x < 70:
            fd.write(" " * (70 - x))

        if branches.has_key(address):
            fd.write("  BR")

        q = ((address % LOCAL_MEMORY_SIZE) / BLOCK_SIZE) 
        assert q in ecb
        if not (q in altmeyer):
            fd.write("  INFEAS?")

        old_address = address    
        fd.write("\n")

    fd.write("\n\n")

    fd.close()
    return (len(ecb), num_ucbs, [ len(x) * 4 for x in regions ])

def ecb_set_to_string(ecb):
    out = []
    for i in xrange(LOCAL_MEMORY_SIZE / BLOCK_SIZE):
        if i in ecb:
            out.append("1")
        else:
            out.append("0")
    return ''.join(out)

def explore_program(instructions, branches, start, stop):
    frontier = [start]
    done = set()

    while len(frontier) != 0:
        address = frontier.pop()
        if address in done:
            continue

        if address == stop:
            continue

        if not instructions.has_key(address):
            print 'jumped out of program at %x' % address
            print call_stack
            sys.exit(1) 

        done.add(address)

        (itype, conditional, target) = branches.get(address, (None, False, None))

        if conditional:
            frontier.append(address + 4)

        if itype == None:
            frontier.append(address + 4)

        elif itype == CALL:
            frontier.append(address + 4)
            frontier.append(target)

        elif itype == RETURN:
            pass

        elif itype == JUMP:
            frontier.append(target)

        else:
            print "unknown itype '%s' at %x" % (itype, address)
            sys.exit(1)

    return done

def get_progs():
    progs = [ x for x in dir(ait_data.ipdp) if not x.startswith("_") ]
    progs.sort()
    return progs

def main():
    progs = get_progs()
    print progs

    out_h = file("spm_data.h", "wt")
    out_h.write("""
#ifndef SPM_DATA
#define SPM_DATA
""")

    bs_time = bs_total = 0
    for name in sorted(globals().keys()):
        if name[0].isupper():
            value = globals()[name]
            if type(value) == float:
                out_h.write("static const double %s = %1.6e;\n" % (name, value))

    out_h.write("""
typedef struct MRTC_Data_struct {
        const char * name;
        double perfect_wcet;
        double cache_wcet;
        double spm_wcet;
        unsigned num_ucbs;
        unsigned num_ecbs;
        unsigned S_spm;
        unsigned num_regions;
        unsigned num_blocks;
        unsigned S_initial_load;
    } MRTC_Data;

static MRTC_Data mrtc_table[] = {
""")

    out_basetasks = file("res/basetasks.tex", "wt")
    out_basetasks.write(r"""
\begin{tabular}{lrrrr}
\hline
Task $\tau_i$ 
        & $C^\execute_i$ (ns)
        & $|\ECB_i|$ 
        & $C^\cache_i$ (ns)
        & $|\UCB_i|$ 
        \\
\hline
""")
    out_spm_data = file("res/spm_data.tex", "wt")
    out_spm_data.write(r"""
\begin{tabular}{lrrrrrr}
\hline
Task $\tau_i$ & $C^\spm_i$ (ns) & $S^\spm_i$ & 
        $\frac{C^\spm_i}{C^\cache_i}$ &
        $\frac{S^\spm_i}{|\ECB_i|}$ &
        $\frac{S^\spm_i}{|\UCB_i|}$ \\
\hline
""")
    discrepancy = file("discrepancy.txt", "wt")
    reject_count = 0
    
    for name in progs:
        assert os.path.exists("binaries/%s.out" % name)

        perfect_wcet = getattr(ait_data.ipdp, name) * CLOCK_CYCLE

        altmeyer_num_ecbs = getattr(ait_data.ecbs, name)
        altmeyer_num_ucbs = getattr(ait_data.ucbs, name)

        if altmeyer_num_ecbs > (LOCAL_MEMORY_SIZE / BLOCK_SIZE):
            print ''
            print 'rejecting %s, too large' % name
            print ''
            reject_count += 1
            continue

        (num_ecbs, num_ucbs, region_sizes) = analyse_ecb_ucb(name)

        altmeyer_num_ecbs = num_ecbs
        altmeyer_num_ucbs = num_ucbs

        cache_wcet = perfect_wcet + (altmeyer_num_ecbs * BRT_CACHE)
        spm_wcet = perfect_wcet 
        total_blocks = 0
        S_spm = 0
        S_initial_load = 0
        print '....', spm_wcet, total_blocks
        for i in xrange(len(region_sizes)):
            size = region_sizes[i]
            if i != (len(region_sizes) - 1):
                size += TRANSITION_CODE_OVERHEAD

            num_blocks = (size + BLOCK_SIZE - 1) / BLOCK_SIZE
            total_blocks += num_blocks
            spm_wcet += fpga_data.Load(num_blocks)
            S_spm = max(S_spm, num_blocks)
            if i == 0:
                S_initial_load = num_blocks
            print '.. .', spm_wcet, total_blocks


        #spm_wcet += fpga_data.Save(S_spm)
        print '... ', spm_wcet, total_blocks
        num_regions = len(region_sizes)

        out_h.write('{"%s", %1.6e, %1.6e, %1.6e, %u, %u, %u, %u, %u, %u},\n' % (
                name, perfect_wcet, cache_wcet, spm_wcet,
                altmeyer_num_ucbs, altmeyer_num_ecbs, 
                S_spm, num_regions, total_blocks,
                S_initial_load))
        out_h.flush()

        x = (perfect_wcet + (num_regions * TRANSITION_WCET_OVERHEAD) +
                (total_blocks * BRT_SPM))
        assert abs(spm_wcet - x) < 1e-8, (spm_wcet, x, name)

        #cache_wcet += CS1
        #spm_wcet += CS1

        out_basetasks.write(
                "%20s &$ %9.0f $&$ %4u $&$ %9.0f $&$ %4u $\\\\\n" % (
                name,
                perfect_wcet * 1e9,
                altmeyer_num_ecbs,
                cache_wcet * 1e9,
                altmeyer_num_ucbs))
        out_spm_data.write(
                "%20s &$ %9.0f $&$ %4u $&$ %1.3f "
                        "$&$ %1.3f $&$ %1.3f $\\\\\n" % (
                name,
                #len(region_sizes),
                spm_wcet * 1e9,
                S_spm,
                spm_wcet / cache_wcet,
                S_spm / float(altmeyer_num_ecbs),
                S_spm / float(altmeyer_num_ucbs)
                ))

        if name == "binarysearch":
            bs_time = perfect_wcet
            bs_total = spm_wcet

        if altmeyer_num_ecbs != num_ecbs:
            discrepancy.write("%s: |ECB| discrepancy %u (Altmeyer) %u (Whitham)\n" % (
                    name, altmeyer_num_ecbs, num_ecbs))

        if altmeyer_num_ucbs != num_ucbs:
            discrepancy.write("%s: |UCB| discrepancy %u (Altmeyer) %u (Whitham)\n" % (
                    name, altmeyer_num_ucbs, num_ucbs))

        altmeyer_cache_wcet = getattr(ait_data.icdp, name) * CLOCK_CYCLE
        whitham_cache_wcet = perfect_wcet + (altmeyer_num_ecbs * BRT_CACHE)
        if abs(altmeyer_cache_wcet - whitham_cache_wcet) > 1e-10:
            discrepancy.write("%s: cache wcet discrepancy %1.6e (Altmeyer) %1.6e (Whitham)\n" % (
                    name, altmeyer_cache_wcet, whitham_cache_wcet))

        

    discrepancy.close()

    out_basetasks.write(r"\hline\end{tabular}")
    out_basetasks.write("\n\n")
    out_basetasks.close()

    out_spm_data.write(r"\hline\end{tabular}")
    out_spm_data.write("\n\n")
    out_spm_data.close()

    out_h.write('};\n\n')
    out_h.write('static const unsigned NUM_MRTC_TASKS = %u;\n\n' % (
            len(progs) - reject_count))
    out_h.write('#endif\n\n\n')
    out_h.close()

    print 'B_CACHE = %1.0f ns' % (B_CACHE * 1e9)
    print 'B_SPM = %1.0f ns' % (B_SPM * 1e9)

    out_params = file("res/params.tex", "wt")
    out_params.write(r"\begin{tabular}{lrl}" + "\n")
    out_params.write(r"\hline" + "\n")
    out_params.write(r"System Parameter & Upper Bound Value & \\" + "\n")
    out_params.write(r"\hline" + "\n")

    eol = r"\\" + "\n"
    out_params.write(r"Clock Period & $%1.0f$ & ns %s" % (
                CLOCK_CYCLE * 1e9, eol))
    out_params.write(r"Block Size & $%u$ & bytes %s" % (BLOCK_SIZE, eol))
    out_params.write(r"Local Memory Size & $%u$ & blocks %s" % (
                LOCAL_MEMORY_SIZE / BLOCK_SIZE, eol))
    out_params.write(r"$\BRT^\cache$ & $%1.0f$ & ns %s" % (BRT_CACHE * 1e9, eol))
    out_params.write(r"$CS^\cto$ & $%1.0f$ & ns %s" % (CS1 * 1e9, eol))
    out_params.write(r"$CS^\from$ & $%1.0f$ & ns %s" % (CS2 * 1e9, eol))
    out_params.write(r"$B^\cache$ & $%1.0f$ & ns %s" % (B_CACHE * 1e9, eol))
    out_params.write(r"$C^\save_i$ & $%1.0fS^\spm_i + %1.0f$ & ns %s" % (
                SAVE_PER_BLOCK * 1e9, 
                (SAVE_OVERHEAD - TRANSITION_WCET_OVERHEAD) * 1e9, eol))
    out_params.write(r"$\BRT^\spm$ & $%1.0f$ & ns %s" % (BRT_SPM * 1e9, eol))
    out_params.write(r"$C^\load_{i,j}$ & $\BRT^\spm S^\spm_{i,j} + %1.0f$"
                                "& ns %s" % (
                #BRT_SPM * 1e9, 
                TRANSITION_WCET_OVERHEAD * 1e9, eol))
    assert (SAVE_OVERHEAD - TRANSITION_WCET_OVERHEAD) > 0
    out_params.write(r"$C^\restore_i$ & $\BRT^\spm S^\spm_i + %1.0f$ & ns %s" % (
                #BRT_SPM * 1e9, 
                RESTORE_OVERHEAD * 1e9, eol))
    out_params.write(r"\hline" + "\n")
    out_params.write(r"\end{tabular}" + "\n")
    out_params.close()

    a = (SAVE_PER_BLOCK + BRT_SPM) * 1e9
    b = (RESTORE_OVERHEAD + SAVE_OVERHEAD - TRANSITION_WCET_OVERHEAD) * 1e9
    out_eqns = file("res/eqns.tex", "wt")
    out_eqns.write(r"""
\newcommand{\egamma}{
\begin{equation}
\begin{split}
    \gamma^\SRPD_{i,j} ~ & = C^\save_j + C^\restore_j \\
                & = %1.0f S^\spm_j + %1.0f
    \label{gammasrpd}
\end{split}
\end{equation}}""" % (a, b))

    out_eqns.write(r"""
\newcommand{\ebbounds}{
\begin{equation}
\begin{split}
    B^\spm_i =~ \max( & \max_{k \in \lp(i)}(%1.0f S^\spm_k + %1.0f), \\
                            & %1.0f S^\spm_i + %1.0f) 
    \label{bspm0} 
\end{split}
\end{equation}}""" % (
            a, (CS1 + SAVE_OVERHEAD) * 1e9,
            BRT_SPM * 1e9, (CS2 + RESTORE_OVERHEAD) * 1e9))
    out_eqns.write(r"""
\newcommand{\ecbounds}{
\begin{equation}
    C^\spm_i = %1.0f |\ECB_i| + %1.0f + C^\execute_i
    \label{cspm0} 
\end{equation}}""" % (BRT_SPM * 1e9,
                        TRANSITION_WCET_OVERHEAD * 1e9))

    def LC(x):
        return 1e9 * ((BRT_SPM * x) + TRANSITION_WCET_OVERHEAD)

    a = (SAVE_PER_BLOCK * 14) + (SAVE_OVERHEAD - TRANSITION_WCET_OVERHEAD)
    a *= 1e9
    out_eqns.write(r"""
\newcommand{\eloadcosts}{
    $C^\load_{i, 1} = %1.0f$,
    $C^\load_{i, 2} = %1.0f$,
    $C^\load_{i, 3} = %1.0f$}""" % (LC(6), LC(14), LC(1)))
    b = LC(6) + LC(14) + LC(1) + ((bs_time) * 1e9)
    out_eqns.write(r"""
\newcommand{\esavecost}{
    $C^\save_i = %1.0f$}""" % (a))
    out_eqns.write(r"""
\newcommand{\etotal}{
    $C^\spm_i = %1.0f$}""" % (b))

    out_eqns.write(r"""
\newcommand{\brtspm}{$%1.0f$}""" % (BRT_SPM * 1e9))
    out_eqns.write(r"""
\newcommand{\brtcache}{$%1.0f$}""" % (BRT_CACHE * 1e9))
    out_eqns.write(r"""
\newcommand{\pratio}{$x = %1.2f$}""" % (BRT_SPM / BRT_CACHE))

    out_eqns.close()
    print 'binarysearch %1.0f %1.0f' % (b, bs_total * 1e9)
    assert int(b) == int(bs_total * 1e9)
    return

if __name__ == "__main__":
    main()
    sys.exit(0)


