


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

library unimacro;
use unimacro.vcomponents.all;

entity logic is
port (
    clk                 : in std_logic;
    rst                 : in std_logic;

    fsl_to_cpu_read	    : in  std_logic;
    fsl_to_cpu_data	    : out std_logic_vector(31 downto 0);
    fsl_to_cpu_control	: out std_logic;
    fsl_to_cpu_exists	: out std_logic;

    fsl_from_cpu_write	: in  std_logic;
    fsl_from_cpu_data	: in  std_logic_vector(31 downto 0);
    fsl_from_cpu_control: in  std_logic;
    fsl_from_cpu_full	: out std_logic;

    fsl_from_mem_read   : out std_logic;
    fsl_from_mem_data   : in  std_logic_vector(31 downto 0);
    fsl_from_mem_control: in  std_logic;
    fsl_from_mem_exists : in  std_logic;

    fsl_to_mem_write    : out std_logic;
    fsl_to_mem_data     : out std_logic_vector(31 downto 0);
    fsl_to_mem_control  : out std_logic;
    fsl_to_mem_full     : in  std_logic;

    db_status           : out std_logic_vector(31 downto 0);
    db_stack_topout     : out std_logic_vector(31 downto 0);
    db_stack_topin      : out std_logic_vector(31 downto 0);
    db_ba_store_in      : out std_logic_vector(31 downto 0);
    db_ba_store_out     : out std_logic_vector(31 downto 0);

    spm_addr1           : in std_logic_vector(31 downto 0);
    spm_to_cpu1         : out std_logic_vector(31 downto 0);
    spm_addr2           : in std_logic_vector(31 downto 0);
    spm_to_cpu2         : out std_logic_vector(31 downto 0)
);
end entity logic;


architecture basic of logic is

    constant SPM_WIDTH      : Natural := 12;
    constant BLOCK_WIDTH    : Natural := 2;

    constant SPM_LEFT       : Natural := SPM_WIDTH - 1;
    constant BLOCK_SIZE     : Natural := 2 ** BLOCK_WIDTH;
    constant BLOCK_LEFT     : Natural := BLOCK_WIDTH - 1;

    type State_Type is (S_INITIAL, S_LOAD, S_RESTORE, S_SAVE);
    type Command_Type is (NO_COMMAND, SET_SPM_INDEX, RESTORE, 
                        SAVE, LOAD, LOAD_AND_SAVE, BLOCKED);

    -- control lines & registers   
    signal state            : State_Type;
    signal command          : Command_Type;
    signal count            : Natural range 0 to BLOCK_SIZE;
    signal spm_index        : std_logic_vector(SPM_LEFT downto 0);
    signal flow             : std_logic;
    signal busy             : std_logic;
    signal dmode            : std_logic_vector(3 downto 0);

    -- internal lines
    signal stack_push       : std_logic;
    signal stack_pop        : std_logic;
    signal stack_topout     : std_logic_vector(31 downto 0);
    signal stack_topin      : std_logic_vector(31 downto 0);
    signal data_store_in    : std_logic_vector(31 downto 0);
    signal data_store_out   : std_logic_vector(31 downto 0);
    signal data_store_we    : std_logic;
    signal ba_store_in      : std_logic_vector(31 downto 0);
    signal ba_store_out     : std_logic_vector(31 downto 0);
    signal stack_x          : std_logic_vector(31 downto 0);
    signal ba_store_we      : std_logic;
    signal stack_ptr        : std_logic_vector(31 downto 0);

    signal memA, memB       : std_logic_vector(31 downto 0);
    signal memC, memD       : std_logic_vector(31 downto 0);

    signal i_status         : std_logic_vector(31 downto 0);
    signal xba_store_we     : std_logic_vector(3 downto 0);

begin
    flow <= fsl_from_mem_exists;

    data_store_in <= fsl_from_mem_data;
    data_store_we <= flow;

    fsl_from_mem_read <= flow;

    command <= 
        NO_COMMAND when (state /= S_INITIAL) else
        NO_COMMAND when (fsl_from_cpu_write = '0') else
        SET_SPM_INDEX when (fsl_from_cpu_control = '0') else
        SAVE when (fsl_from_cpu_data = x"00000001") else
        BLOCKED when (fsl_to_mem_full = '1') else
        RESTORE when (fsl_from_cpu_data = x"00000000") else
        LOAD when (fsl_from_cpu_data(0) = '0') else
        LOAD_AND_SAVE;
     
    fsl_to_mem_control <= '0';

    stack_topin <= ba_store_out;

    fsl_from_cpu_full <= busy;
    fsl_to_cpu_exists <= not busy;
    fsl_to_cpu_control <= '0';

    process (clk, rst) is
    begin
        if rst = '1' then
            state <= S_INITIAL;
            spm_index <= (others => '0');
            count <= 0;
            stack_x <= (others => '0');

        elsif (clk'event and clk = '1') then
            stack_pop <= '0';
            stack_push <= '0';
            ba_store_we <= '0';

            case state is
            when S_INITIAL =>
                busy <= '0';
                case command is
                when SET_SPM_INDEX =>
                    spm_index <= (others => '0');
                    spm_index(SPM_LEFT downto BLOCK_WIDTH) <= 
                        fsl_from_cpu_data(SPM_LEFT + 2 downto BLOCK_WIDTH + 2);
                    dmode <= fsl_from_cpu_data(31 downto 28);

                when RESTORE =>
                    ba_store_in <= stack_topout;
                    stack_pop <= '1';
                    state <= S_RESTORE;
                    count <= BLOCK_SIZE;
                    ba_store_we <= '1';
                    busy <= '1';

                when LOAD|LOAD_AND_SAVE =>
                    ba_store_in <= fsl_from_cpu_data;
                    state <= S_LOAD;
                    count <= BLOCK_SIZE;
                    ba_store_we <= '1';
                    busy <= '1';

                    if command = LOAD_AND_SAVE then
                        stack_push <= '1';
                    end if;

                when SAVE =>
                    stack_push <= '1';
                    if conv_integer(spm_index) = 0 then
                        -- Exactly one to stack
                        null;
                    else
                        -- More the one to stack
                        state <= S_SAVE;
                        spm_index <= spm_index - BLOCK_SIZE;
                        busy <= '1';
                    end if;

                when others =>
                    null;
                end case;

            when S_SAVE =>
                stack_push <= '1';
                if conv_integer(spm_index) = 0 then
                    -- Now stacking final element
                    state <= S_INITIAL;
                    busy <= '0';
                else
                    -- More elements remain
                    spm_index <= spm_index - BLOCK_SIZE;
                end if; 

            when S_RESTORE|S_LOAD =>
                if flow = '1' then
                    spm_index <= spm_index + 1;
                    count <= count - 1;
                    if count = 1 then
                        state <= S_INITIAL;
                        busy <= '0';
                        if state = S_LOAD then
                            spm_index <= spm_index + 1 - (BLOCK_SIZE * 2);
                        end if;
                    end if;
                end if;
            end case;
            stack_x <= stack_topout or stack_x;
        end if;
    end process;

    process (command, fsl_from_cpu_data, stack_topout) is
    begin
        fsl_to_mem_data <= fsl_from_cpu_data;
        fsl_to_mem_write <= '0';
        case command is
        when RESTORE =>
            fsl_to_mem_data <= stack_topout;
            fsl_to_mem_write <= '1';
        when LOAD|LOAD_AND_SAVE =>
            fsl_to_mem_write <= '1';
        when others =>
            null;
        end case;
        fsl_to_mem_data(BLOCK_LEFT downto 0) <= (others => '0');
    end process;
                            

    s : entity stack 
            port map (
                clk => clk,
                rst => rst,
                push => stack_push,
                pop => stack_pop,
                topout => stack_topout,
                db_ptr => stack_ptr,
                topin => stack_topin);


    lane_gen : for lane in 0 to 3
    generate
        b : block
            constant right  : Natural := 8 * lane;
            constant left   : Natural := right + 7;
            signal xwe      : std_logic_vector(3 downto 0);
        begin
            xwe <= (others => data_store_we);
            data_store : BRAM_TDP_MACRO
                generic map (
                    BRAM_SIZE => "36Kb",
                    WRITE_WIDTH_A => 8,
                    READ_WIDTH_A => 8,
                    WRITE_WIDTH_B => 8,
                    READ_WIDTH_B => 8)
                port map (
                    diA => data_store_in(left downto right),
                    addrA => spm_index,
                    weA => xwe,
                    enA => '1',
                    rstA => rst,
                    regceA => '0',
                    clkA => clk,
                    doA => data_store_out(left downto right),

                    diB => x"00",
                    addrB => spm_addr1(SPM_LEFT + 2 downto 2),
                    weB => "0",
                    enB => '1',
                    rstB => rst,
                    regceB => '0',
                    clkB => clk,
                    doB => spm_to_cpu1(left downto right));
        end block b;
    end generate lane_gen;
    
    xba_store_we <= (others => ba_store_we);
    ba_store : BRAM_SINGLE_MACRO
            generic map (
                BRAM_SIZE => "36Kb",
                WRITE_WIDTH => 32,
                READ_WIDTH => 32)
            port map (
                di => ba_store_in,
                addr => spm_index(SPM_LEFT downto BLOCK_WIDTH),
                we => xba_store_we,
                en => '1',
                regce => '0',
                rst => rst,
                clk => clk,
                do => ba_store_out);

    debug : process(state, command, spm_index, flow, rst, busy) is
    begin
        i_status <= (others => '0');
        case state is
        when S_INITIAL => i_status(31 downto 28) <= "0001";
        when S_LOAD    => i_status(31 downto 28) <= "0010";
        when S_RESTORE => i_status(31 downto 28) <= "0100";
        when S_SAVE =>    i_status(31 downto 28) <= "0101";
        end case;
        case command is
        when NO_COMMAND =>      i_status(27 downto 24) <= "0001";
        when SET_SPM_INDEX =>   i_status(27 downto 24) <= "0010";
        when RESTORE =>         i_status(27 downto 24) <= "0011";
        when SAVE =>            i_status(27 downto 24) <= "0100";
        when LOAD =>            i_status(27 downto 24) <= "0101";
        when LOAD_AND_SAVE =>   i_status(27 downto 24) <= "0110";
        when BLOCKED =>         i_status(27 downto 24) <= "0111";
        end case;
        i_status(23) <= flow;
        i_status(21) <= rst;
        i_status(22) <= fsl_to_mem_full;
        i_status(20) <= busy;
        i_status(19) <= stack_push;
        i_status(18) <= stack_pop;
        i_status(17) <= ba_store_we;
        i_status(SPM_LEFT downto 0) <= spm_index;
    end process debug;

    db_status <= i_status;
    fsl_to_cpu_data <= 
            i_status when dmode = "0000" else
            stack_topout when dmode = "0001" else
            stack_topin when dmode = "0010" else
            data_store_in when dmode = "0011" else
            data_store_out when dmode = "0100" else
            ba_store_in when dmode = "0101" else
            ba_store_out when dmode = "0110" else
            stack_ptr when dmode = "0111" else
            memA when dmode = "1000" else
            memB when dmode = "1001" else
            memC when dmode = "1010" else
            memD when dmode = "1011" else
            stack_x;

    process (clk) is
    begin
        if clk = '1' and clk'event then
            if data_store_we = '1' then
                case conv_integer(spm_index) is
                when 0 => memA <= data_store_in;
                when 1 => memB <= data_store_in;
                when 2 => memC <= data_store_in;
                when 3 => memD <= data_store_in;
                when others => null;
                end case;
            end if;
        end if;
    end process;

    db_stack_topout <= stack_topout;
    db_stack_topin <= stack_topin;
    db_ba_store_in <= ba_store_in;
    db_ba_store_out <= ba_store_out;

end architecture basic;



